@@ -26,7 +26,6 @@
  */
 #define DUK__ENUM_START_INDEX  2
 
-#if 0
 /* Current implementation suffices for ES2015 for now because there's no symbol
  * sorting, so commented out for now.
  */
@@ -35,55 +34,72 @@
  *  Helper to sort enumeration keys using a callback for pairwise duk_hstring
  *  comparisons.  The keys are in the enumeration object entry part, starting
  *  from DUK__ENUM_START_INDEX, and the entry part is dense.  Entry part values
- *  are all "true", e.g. "1" -> true, "3" -> true, "foo" -> true "2" -> true,
+ *  are all "true", e.g. "1" -> true, "3" -> true, "foo" -> true, "2" -> true,
  *  so it suffices to just switch keys without switching values.
  *
+ *  ES2015 [[OwnPropertyKeys]] enumeration order for ordinary objects:
+ *  (1) array indices in ascending order,
+ *  (2) non-array-index keys in insertion order, and
+ *  (3) symbols in insertion order.
+ *  http://www.ecma-international.org/ecma-262/6.0/#sec-ordinary-object-internal-methods-and-internal-slots-ownpropertykeys.
+ *
+ *  This rule is applied to "own properties" at each inheritance level;
+ *  non-duplicate parent keys always follow child keys.  For example,
+ *  an inherited array index will enumerate -after- a symbol in the
+ *  child.
+ *
  *  Insertion sort is used because (1) it's simple and compact, (2) works
  *  in-place, (3) minimizes operations if data is already nearly sorted,
  *  (4) doesn't reorder elements considered equal.
  *  http://en.wikipedia.org/wiki/Insertion_sort
  */
 
-typedef duk_bool_t (*duk__sort_compare_fn)(duk_hstring *a, duk_hstring *b, duk_uarridx_t val_b);
+/* Sort key, must hold array indices, "not array index" marker, and one more
+ * higher value for symbols.
+ */
+#if !defined(DUK_USE_SYMBOL_BUILTIN)
+typedef duk_uint32_t duk__sort_key_t;
+#elif defined(DUK_USE_64BIT_OPS)
+typedef duk_uint64_t duk__sort_key_t;
+#else
+typedef duk_double_t duk__sort_key_t;
+#endif
 
-DUK_LOCAL duk_bool_t duk__sort_compare_es6(duk_hstring *a, duk_hstring *b, duk_uarridx_t val_b) {
-	duk_uarridx_t val_a;
+/* Get sort key for a duk_hstring. */
+DUK_LOCAL duk__sort_key_t duk__hstring_sort_key(duk_hstring *x) {
+	duk__sort_key_t val;
+
+	/* For array indices [0,0xfffffffe] use the array index as is.
+	 * For strings, use 0xffffffff, the marker 'arridx' already in
+	 * duk_hstring.  For symbols, any value above 0xffffffff works,
+	 * as long as it is the same for all symbols; currently just add
+	 * the masked flag field into the arridx temporary.
+	 */
+	DUK_ASSERT(x != NULL);
+	DUK_ASSERT(!DUK_HSTRING_HAS_SYMBOL(x) || DUK_HSTRING_GET_ARRIDX_FAST(x) == DUK_HSTRING_NO_ARRAY_INDEX);
+
+	val = (duk__sort_key_t) DUK_HSTRING_GET_ARRIDX_FAST(x);
+
+#if defined(DUK_USE_SYMBOL_BUILTIN)
+	val = val + (duk__sort_key_t) (DUK_HEAPHDR_GET_FLAGS_RAW((duk_heaphdr *) x) & DUK_HSTRING_FLAG_SYMBOL);
+#endif
+
+	return (duk__sort_key_t) val;
+}
+
+/* Insert element 'b' after element 'a'? */
+DUK_LOCAL duk_bool_t duk__sort_compare_es6(duk_hstring *a, duk_hstring *b, duk__sort_key_t val_b) {
+	duk__sort_key_t val_a;
 
 	DUK_ASSERT(a != NULL);
 	DUK_ASSERT(b != NULL);
 	DUK_UNREF(b);  /* Not actually needed now, val_b suffices. */
 
-	/* ES2015 [[OwnPropertyKeys]] enumeration order for ordinary objects:
-	 * (1) array indices in ascending order, (2) non-array-index keys in
-	 * insertion order, symbols in insertion order:
-	 * http://www.ecma-international.org/ecma-262/6.0/#sec-ordinary-object-internal-methods-and-internal-slots-ownpropertykeys.
-	 *
-	 * This rule is applied to "own properties" at each inheritance level;
-	 * non-duplicate parent keys always follow child keys.  For example,
-	 * an inherited array index will enumerate -after- a symbol in the
-	 * child.
-	 */
+	val_a = duk__hstring_sort_key(a);
 
-	val_a = DUK_HSTRING_GET_ARRIDX_FAST(a);
-
-	if (val_b < val_a) {
-		/* Covers:
-		 *   - Both keys are array indices and a > b: don't insert here.
-		 *   - 'b' is array index, 'a' is not: don't insert here.
-		 */
+	if (val_a > val_b) {
 		return 0;
 	} else {
-		/* Covers:
-		 *   val_a < val_b where:
-		 *   - Both keys are array indices and a < b: insert here.
-		 *   - 'a' is array index, 'b' is not: insert here.
-		 *   val_a == val_b where:
-		 *   - Both keys are array indices and a == b: insert here
-		 *     (shouldn't actually happen, can't have non-duplicate
-		 *     identical array index keys).
-		 *   - Neither key is an array index: insert here, keeps key
-		 *     order regardless of the keys themselves.
-		 */
 		return 1;
 	}
 }
@@ -106,7 +122,7 @@ DUK_LOCAL void duk__sort_enum_keys_es6(duk_hthread *thr, duk_hobject *h_obj, duk
 	for (idx = idx_start + 1; idx < idx_end; idx++) {
 		duk_hstring *h_curr;
 		duk_int_fast32_t idx_insert;
-		duk_uarridx_t val_curr;
+		duk__sort_key_t val_curr;
 
 		h_curr = keys[idx];
 		DUK_ASSERT(h_curr != NULL);
@@ -116,16 +132,12 @@ DUK_LOCAL void duk__sort_enum_keys_es6(duk_hthread *thr, duk_hobject *h_obj, duk
 		 * (and optimized for) case.
 		 */
 
-		val_curr = DUK_HSTRING_GET_ARRIDX_FAST(h_curr);  /* Remains same during scanning. */
+		val_curr = duk__hstring_sort_key(h_curr);  /* Remains same during scanning. */
 		for (idx_insert = idx - 1; idx_insert >= idx_start; idx_insert--) {
 			duk_hstring *h_insert;
 			h_insert = keys[idx_insert];
 			DUK_ASSERT(h_insert != NULL);
 
-			/* XXX: fixed callback rather than a callback argument; only
-			 * one argument used and using a callback argument doesn't
-			 * cause e.g. gcc to inline the callback.
-			 */
 			if (duk__sort_compare_es6(h_insert, h_curr, val_curr)) {
 				break;
 			}
@@ -147,90 +159,13 @@ DUK_LOCAL void duk__sort_enum_keys_es6(duk_hthread *thr, duk_hobject *h_obj, duk
 		 * are very often in order already.
 		 */
 		if (idx != idx_insert) {
-			DUK_MEMMOVE((void *) (keys + idx_insert + 1),
+			duk_memmove((void *) (keys + idx_insert + 1),
 			            (const void *) (keys + idx_insert),
-			            (size_t) ((idx - idx_insert) * sizeof(duk_hstring *)));
+			            ((size_t) (idx - idx_insert) * sizeof(duk_hstring *)));
 			keys[idx_insert] = h_curr;
 		}
 	}
 }
-#endif  /* disabled */
-
-/*
- *  Helper to sort keys into ES2015 [[OwnPropertyKeys]] enumeration order:
- *  array keys in ascending order first, followed by keys in insertion
- *  order, followed by symbols in insertion order (not handled here).
- *  Insertion sort based.
- *
- *  This algorithm nominally sorts array indices, but because the "no array
- *  index" marker is higher than any array index, non-array-index keys are
- *  sorted after array indices.  Non-array-index keys are also considered
- *  equal for sorting which means that their order is kept as is, so the end
- *  result matches ES2015 [[OwnPropertyKeys]].
- *
- *  Insertion sort is used because (1) it's simple and compact, (2) works
- *  in-place, (3) minimizes operations if data is already nearly sorted,
- *  (4) doesn't reorder elements considered equal.
- *  http://en.wikipedia.org/wiki/Insertion_sort
- */
-
-DUK_LOCAL void duk__sort_enum_keys_es6(duk_hthread *thr, duk_hobject *h_obj, duk_int_fast32_t idx_start, duk_int_fast32_t idx_end) {
-	duk_hstring **keys;
-	duk_hstring **p_curr, **p_insert, **p_end;
-	duk_hstring *h_curr;
-	duk_uarridx_t val_highest, val_curr, val_insert;
-
-	DUK_ASSERT(h_obj != NULL);
-	DUK_ASSERT(idx_start >= DUK__ENUM_START_INDEX);
-	DUK_ASSERT(idx_end >= idx_start);
-	DUK_UNREF(thr);
-
-	if (idx_end <= idx_start + 1) {
-		return;  /* Zero or one element(s). */
-	}
-
-	keys = DUK_HOBJECT_E_GET_KEY_BASE(thr->heap, h_obj);
-	p_curr = keys + idx_start;
-	val_highest = DUK_HSTRING_GET_ARRIDX_SLOW(*p_curr);
-	for (p_curr++, p_end = keys + idx_end; p_curr < p_end; p_curr++) {
-		DUK_ASSERT(*p_curr != NULL);
-		val_curr = DUK_HSTRING_GET_ARRIDX_SLOW(*p_curr);
-
-		if (val_curr >= val_highest) {
-			val_highest = val_curr;
-			continue;
-		}
-
-		/* Needs to be inserted; scan backwards, since we optimize
-		 * for the case where elements are nearly in order.
-		 */
-
-		p_insert = p_curr;
-		for (;;) {
-			p_insert--;  /* Start from p_curr - 1. */
-			val_insert = DUK_HSTRING_GET_ARRIDX_SLOW(*p_insert);
-			if (val_insert < val_curr) {
-				p_insert++;
-				break;
-			}
-			if (p_insert == keys + idx_start) {
-				break;
-			}
-		}
-
-		/*        .-- p_insert   .-- p_curr
-		 *        v              v
-		 *  | ... | insert | ... | curr
-		 */
-
-		h_curr = *p_curr;
-		DUK_MEMMOVE((void *) (p_insert + 1),
-		            (const void *) p_insert,
-		            (size_t) ((p_curr - p_insert) * sizeof(duk_hstring *)));
-		*p_insert = h_curr;
-		/* keep val_highest */
-	}
-}
 
 /*
  *  Create an internal enumerator object E, which has its keys ordered
@@ -241,21 +176,20 @@ DUK_LOCAL void duk__sort_enum_keys_es6(duk_hthread *thr, duk_hobject *h_obj, duk
  *  scan would be needed to eliminate duplicates found in the prototype chain.
  */
 
-DUK_LOCAL void duk__add_enum_key(duk_context *ctx, duk_hstring *k) {
+DUK_LOCAL void duk__add_enum_key(duk_hthread *thr, duk_hstring *k) {
 	/* 'k' may be unreachable on entry so must push without any
 	 * potential for GC.
 	 */
-	duk_push_hstring(ctx, k);
-	duk_push_true(ctx);
-	duk_put_prop(ctx, -3);
+	duk_push_hstring(thr, k);
+	duk_push_true(thr);
+	duk_put_prop(thr, -3);
 }
 
-DUK_LOCAL void duk__add_enum_key_stridx(duk_context *ctx, duk_small_uint_t stridx) {
-	duk__add_enum_key(ctx, DUK_HTHREAD_GET_STRING((duk_hthread *) ctx, stridx));
+DUK_LOCAL void duk__add_enum_key_stridx(duk_hthread *thr, duk_small_uint_t stridx) {
+	duk__add_enum_key(thr, DUK_HTHREAD_GET_STRING(thr, stridx));
 }
 
-DUK_INTERNAL void duk_hobject_enumerator_create(duk_context *ctx, duk_small_uint_t enum_flags) {
-	duk_hthread *thr = (duk_hthread *) ctx;
+DUK_INTERNAL void duk_hobject_enumerator_create(duk_hthread *thr, duk_small_uint_t enum_flags) {
 	duk_hobject *enum_target;
 	duk_hobject *curr;
 	duk_hobject *res;
@@ -267,13 +201,13 @@ DUK_INTERNAL void duk_hobject_enumerator_create(duk_context *ctx, duk_small_uint
 	duk_uint_fast32_t i, len;  /* used for array, stack, and entry indices */
 	duk_uint_fast32_t sort_start_index;
 
-	DUK_ASSERT(ctx != NULL);
+	DUK_ASSERT(thr != NULL);
 
-	enum_target = duk_require_hobject(ctx, -1);
+	enum_target = duk_require_hobject(thr, -1);
 	DUK_ASSERT(enum_target != NULL);
 
-	duk_push_bare_object(ctx);
-	res = duk_known_hobject(ctx, -1);
+	duk_push_bare_object(thr);
+	res = duk_known_hobject(thr, -1);
 
 	/* [enum_target res] */
 
@@ -282,12 +216,12 @@ DUK_INTERNAL void duk_hobject_enumerator_create(duk_context *ctx, duk_small_uint
 	 * enumeration result comes from a proxy trap as there is no
 	 * real object to check against.
 	 */
-	duk_push_hobject(ctx, enum_target);
-	duk_put_prop_stridx_short(ctx, -2, DUK_STRIDX_INT_TARGET);
+	duk_push_hobject(thr, enum_target);
+	duk_put_prop_stridx_short(thr, -2, DUK_STRIDX_INT_TARGET);  /* Target is bare, plain put OK. */
 
 	/* Initialize index so that we skip internal control keys. */
-	duk_push_int(ctx, DUK__ENUM_START_INDEX);
-	duk_put_prop_stridx_short(ctx, -2, DUK_STRIDX_INT_NEXT);
+	duk_push_int(thr, DUK__ENUM_START_INDEX);
+	duk_put_prop_stridx_short(thr, -2, DUK_STRIDX_INT_NEXT);  /* Target is bare, plain put OK. */
 
 	/*
 	 *  Proxy object handling
@@ -297,8 +231,7 @@ DUK_INTERNAL void duk_hobject_enumerator_create(duk_context *ctx, duk_small_uint
 	if (DUK_LIKELY((enum_flags & DUK_ENUM_NO_PROXY_BEHAVIOR) != 0)) {
 		goto skip_proxy;
 	}
-	if (DUK_LIKELY(!duk_hobject_proxy_check(thr,
-	                                        enum_target,
+	if (DUK_LIKELY(!duk_hobject_proxy_check(enum_target,
 	                                        &h_proxy_target,
 	                                        &h_proxy_handler))) {
 		goto skip_proxy;
@@ -310,8 +243,8 @@ DUK_INTERNAL void duk_hobject_enumerator_create(duk_context *ctx, duk_small_uint
 	 * has been obsoleted and "ownKeys" is used instead.
 	 */
 	DUK_DDD(DUK_DDDPRINT("proxy enumeration"));
-	duk_push_hobject(ctx, h_proxy_handler);
-	if (!duk_get_prop_stridx_short(ctx, -1, DUK_STRIDX_OWN_KEYS)) {
+	duk_push_hobject(thr, h_proxy_handler);
+	if (!duk_get_prop_stridx_short(thr, -1, DUK_STRIDX_OWN_KEYS)) {
 		/* No need to replace the 'enum_target' value in stack, only the
 		 * enum_target reference.  This also ensures that the original
 		 * enum target is reachable, which keeps the proxy and the proxy
@@ -321,38 +254,38 @@ DUK_INTERNAL void duk_hobject_enumerator_create(duk_context *ctx, duk_small_uint
 		DUK_DDD(DUK_DDDPRINT("h_proxy_target=%!O", (duk_heaphdr *) h_proxy_target));
 		enum_target = h_proxy_target;
 
-		duk_push_hobject(ctx, enum_target);  /* -> [ ... enum_target res handler undefined target ] */
-		duk_put_prop_stridx_short(ctx, -4, DUK_STRIDX_INT_TARGET);
+		duk_push_hobject(thr, enum_target);  /* -> [ ... enum_target res handler undefined target ] */
+		duk_put_prop_stridx_short(thr, -4, DUK_STRIDX_INT_TARGET);  /* Target is bare, plain put OK. */
 
-		duk_pop_2(ctx);  /* -> [ ... enum_target res ] */
+		duk_pop_2(thr);  /* -> [ ... enum_target res ] */
 		goto skip_proxy;
 	}
 
 	/* [ ... enum_target res handler trap ] */
-	duk_insert(ctx, -2);
-	duk_push_hobject(ctx, h_proxy_target);    /* -> [ ... enum_target res trap handler target ] */
-	duk_call_method(ctx, 1 /*nargs*/);        /* -> [ ... enum_target res trap_result ] */
-	h_trap_result = duk_require_hobject(ctx, -1);
+	duk_insert(thr, -2);
+	duk_push_hobject(thr, h_proxy_target);    /* -> [ ... enum_target res trap handler target ] */
+	duk_call_method(thr, 1 /*nargs*/);        /* -> [ ... enum_target res trap_result ] */
+	h_trap_result = duk_require_hobject(thr, -1);
 	DUK_UNREF(h_trap_result);
 
-	duk_proxy_ownkeys_postprocess(ctx, h_proxy_target, enum_flags);
+	duk_proxy_ownkeys_postprocess(thr, h_proxy_target, enum_flags);
 	/* -> [ ... enum_target res trap_result keys_array ] */
 
 	/* Copy cleaned up trap result keys into the enumerator object. */
 	/* XXX: result is a dense array; could make use of that. */
-	DUK_ASSERT(duk_is_array(ctx, -1));
-	len = (duk_uint_fast32_t) duk_get_length(ctx, -1);
+	DUK_ASSERT(duk_is_array(thr, -1));
+	len = (duk_uint_fast32_t) duk_get_length(thr, -1);
 	for (i = 0; i < len; i++) {
-		(void) duk_get_prop_index(ctx, -1, i);
-		DUK_ASSERT(duk_is_string(ctx, -1));  /* postprocess cleaned up */
+		(void) duk_get_prop_index(thr, -1, (duk_uarridx_t) i);
+		DUK_ASSERT(duk_is_string(thr, -1));  /* postprocess cleaned up */
 		/* [ ... enum_target res trap_result keys_array val ] */
-		duk_push_true(ctx);
+		duk_push_true(thr);
 		/* [ ... enum_target res trap_result keys_array val true ] */
-		duk_put_prop(ctx, -5);
+		duk_put_prop(thr, -5);
 	}
 	/* [ ... enum_target res trap_result keys_array ] */
-	duk_pop_2(ctx);
-	duk_remove_m2(ctx);
+	duk_pop_2(thr);
+	duk_remove_m2(thr);
 
 	/* [ ... res ] */
 
@@ -376,6 +309,7 @@ DUK_INTERNAL void duk_hobject_enumerator_create(duk_context *ctx, duk_small_uint
 #if !defined(DUK_USE_PREFER_SIZE)
 		duk_bool_t need_sort = 0;
 #endif
+		duk_bool_t cond;
 
 		/* Enumeration proceeds by inheritance level.  Virtual
 		 * properties need to be handled specially, followed by
@@ -402,10 +336,12 @@ DUK_INTERNAL void duk_hobject_enumerator_create(duk_context *ctx, duk_small_uint
 		 */
 
 #if defined(DUK_USE_BUFFEROBJECT_SUPPORT)
-		if (DUK_HOBJECT_HAS_EXOTIC_STRINGOBJ(curr) || DUK_HOBJECT_IS_BUFOBJ(curr)) {
+		cond = DUK_HOBJECT_HAS_EXOTIC_STRINGOBJ(curr) || DUK_HOBJECT_IS_BUFOBJ(curr);
 #else
-		if (DUK_HOBJECT_HAS_EXOTIC_STRINGOBJ(curr)) {
+		cond = DUK_HOBJECT_HAS_EXOTIC_STRINGOBJ(curr);
 #endif
+		cond = cond && !(enum_flags & DUK_ENUM_EXCLUDE_STRINGS);
+		if (cond) {
 			duk_bool_t have_length = 1;
 
 			/* String and buffer enumeration behavior is identical now,
@@ -445,10 +381,10 @@ DUK_INTERNAL void duk_hobject_enumerator_create(duk_context *ctx, duk_small_uint
 				/* This is a bit fragile: the string is not
 				 * reachable until it is pushed by the helper.
 				 */
-				k = duk_heap_strtable_intern_u32_checked(thr, i);
+				k = duk_heap_strtable_intern_u32_checked(thr, (duk_uint32_t) i);
 				DUK_ASSERT(k);
 
-				duk__add_enum_key(ctx, k);
+				duk__add_enum_key(thr, k);
 
 				/* [enum_target res] */
 			}
@@ -459,11 +395,7 @@ DUK_INTERNAL void duk_hobject_enumerator_create(duk_context *ctx, duk_small_uint
 			 */
 
 			if (have_length && (enum_flags & DUK_ENUM_INCLUDE_NONENUMERABLE)) {
-				duk__add_enum_key_stridx(ctx, DUK_STRIDX_LENGTH);
-			}
-		} else if (DUK_HOBJECT_HAS_EXOTIC_DUKFUNC(curr)) {
-			if (enum_flags & DUK_ENUM_INCLUDE_NONENUMERABLE) {
-				duk__add_enum_key_stridx(ctx, DUK_STRIDX_LENGTH);
+				duk__add_enum_key_stridx(thr, DUK_STRIDX_LENGTH);
 			}
 		}
 
@@ -471,26 +403,29 @@ DUK_INTERNAL void duk_hobject_enumerator_create(duk_context *ctx, duk_small_uint
 		 *  Array part
 		 */
 
-		for (i = 0; i < (duk_uint_fast32_t) DUK_HOBJECT_GET_ASIZE(curr); i++) {
-			duk_hstring *k;
-			duk_tval *tv;
+		cond = !(enum_flags & DUK_ENUM_EXCLUDE_STRINGS);
+		if (cond) {
+			for (i = 0; i < (duk_uint_fast32_t) DUK_HOBJECT_GET_ASIZE(curr); i++) {
+				duk_hstring *k;
+				duk_tval *tv;
 
-			tv = DUK_HOBJECT_A_GET_VALUE_PTR(thr->heap, curr, i);
-			if (DUK_TVAL_IS_UNUSED(tv)) {
-				continue;
+				tv = DUK_HOBJECT_A_GET_VALUE_PTR(thr->heap, curr, i);
+				if (DUK_TVAL_IS_UNUSED(tv)) {
+					continue;
+				}
+				k = duk_heap_strtable_intern_u32_checked(thr, (duk_uint32_t) i);  /* Fragile reachability. */
+				DUK_ASSERT(k);
+
+				duk__add_enum_key(thr, k);
+
+				/* [enum_target res] */
 			}
-			k = duk_heap_strtable_intern_u32_checked(thr, i);  /* Fragile reachability. */
-			DUK_ASSERT(k);
 
-			duk__add_enum_key(ctx, k);
-
-			/* [enum_target res] */
-		}
-
-		if (DUK_HOBJECT_HAS_EXOTIC_ARRAY(curr)) {
-			/* Array .length comes after numeric indices. */
-			if (enum_flags & DUK_ENUM_INCLUDE_NONENUMERABLE) {
-				duk__add_enum_key_stridx(ctx, DUK_STRIDX_LENGTH);
+			if (DUK_HOBJECT_HAS_EXOTIC_ARRAY(curr)) {
+				/* Array .length comes after numeric indices. */
+				if (enum_flags & DUK_ENUM_INCLUDE_NONENUMERABLE) {
+					duk__add_enum_key_stridx(thr, DUK_STRIDX_LENGTH);
+				}
 			}
 		}
 
@@ -517,6 +452,9 @@ DUK_INTERNAL void duk_hobject_enumerator_create(duk_context *ctx, duk_small_uint
 				if (!(enum_flags & DUK_ENUM_INCLUDE_SYMBOLS)) {
 					continue;
 				}
+#if !defined(DUK_USE_PREFER_SIZE)
+				need_sort = 1;
+#endif
 			} else {
 				DUK_ASSERT(!DUK_HSTRING_HAS_HIDDEN(k));  /* would also have symbol flag */
 				if (enum_flags & DUK_ENUM_EXCLUDE_STRINGS) {
@@ -540,7 +478,7 @@ DUK_INTERNAL void duk_hobject_enumerator_create(duk_context *ctx, duk_small_uint
 			DUK_ASSERT(DUK_HOBJECT_E_SLOT_IS_ACCESSOR(thr->heap, curr, i) ||
 			           !DUK_TVAL_IS_UNUSED(&DUK_HOBJECT_E_GET_VALUE_PTR(thr->heap, curr, i)->v));
 
-			duk__add_enum_key(ctx, k);
+			duk__add_enum_key(thr, k);
 
 			/* [enum_target res] */
 		}
@@ -563,11 +501,11 @@ DUK_INTERNAL void duk_hobject_enumerator_create(duk_context *ctx, duk_small_uint
 
 		if (!(enum_flags & DUK_ENUM_SORT_ARRAY_INDICES)) {
 #if defined(DUK_USE_PREFER_SIZE)
-			duk__sort_enum_keys_es6(thr, res, sort_start_index, sort_end_index);
+			duk__sort_enum_keys_es6(thr, res, (duk_int_fast32_t) sort_start_index, (duk_int_fast32_t) sort_end_index);
 #else
 			if (need_sort) {
 				DUK_DDD(DUK_DDDPRINT("need to sort"));
-				duk__sort_enum_keys_es6(thr, res, sort_start_index, sort_end_index);
+				duk__sort_enum_keys_es6(thr, res, (duk_int_fast32_t) sort_start_index, (duk_int_fast32_t) sort_end_index);
 			} else {
 				DUK_DDD(DUK_DDDPRINT("no need to sort"));
 			}
@@ -585,7 +523,7 @@ DUK_INTERNAL void duk_hobject_enumerator_create(duk_context *ctx, duk_small_uint
 
 	/* [enum_target res] */
 
-	duk_remove_m2(ctx);
+	duk_remove_m2(thr);
 
 	/* [res] */
 
@@ -600,7 +538,7 @@ DUK_INTERNAL void duk_hobject_enumerator_create(duk_context *ctx, duk_small_uint
 		/* Sort to ES2015 order which works for pure array incides but
 		 * also for mixed keys.
 		 */
-		duk__sort_enum_keys_es6(thr, res, DUK__ENUM_START_INDEX, DUK_HOBJECT_GET_ENEXT(res));
+		duk__sort_enum_keys_es6(thr, res, (duk_int_fast32_t) DUK__ENUM_START_INDEX, (duk_int_fast32_t) DUK_HOBJECT_GET_ENEXT(res));
 	}
 
 #if defined(DUK_USE_ES6_PROXY)
@@ -609,7 +547,7 @@ DUK_INTERNAL void duk_hobject_enumerator_create(duk_context *ctx, duk_small_uint
 	/* compact; no need to seal because object is internal */
 	duk_hobject_compact_props(thr, res);
 
-	DUK_DDD(DUK_DDDPRINT("created enumerator object: %!iT", (duk_tval *) duk_get_tval(ctx, -1)));
+	DUK_DDD(DUK_DDDPRINT("created enumerator object: %!iT", (duk_tval *) duk_get_tval(thr, -1)));
 }
 
 /*
@@ -620,24 +558,23 @@ DUK_INTERNAL void duk_hobject_enumerator_create(duk_context *ctx, duk_small_uint
  *
  *  Returns zero without pushing anything on the stack otherwise.
  */
-DUK_INTERNAL duk_bool_t duk_hobject_enumerator_next(duk_context *ctx, duk_bool_t get_value) {
-	duk_hthread *thr = (duk_hthread *) ctx;
+DUK_INTERNAL duk_bool_t duk_hobject_enumerator_next(duk_hthread *thr, duk_bool_t get_value) {
 	duk_hobject *e;
 	duk_hobject *enum_target;
 	duk_hstring *res = NULL;
 	duk_uint_fast32_t idx;
 	duk_bool_t check_existence;
 
-	DUK_ASSERT(ctx != NULL);
+	DUK_ASSERT(thr != NULL);
 
 	/* [... enum] */
 
-	e = duk_require_hobject(ctx, -1);
+	e = duk_require_hobject(thr, -1);
 
 	/* XXX use get tval ptr, more efficient */
-	duk_get_prop_stridx_short(ctx, -1, DUK_STRIDX_INT_NEXT);
-	idx = (duk_uint_fast32_t) duk_require_uint(ctx, -1);
-	duk_pop(ctx);
+	duk_get_prop_stridx_short(thr, -1, DUK_STRIDX_INT_NEXT);
+	idx = (duk_uint_fast32_t) duk_require_uint(thr, -1);
+	duk_pop(thr);
 	DUK_DDD(DUK_DDDPRINT("enumeration: index is: %ld", (long) idx));
 
 	/* Enumeration keys are checked against the enumeration target (to see
@@ -645,18 +582,18 @@ DUK_INTERNAL duk_bool_t duk_hobject_enumerator_next(duk_context *ctx, duk_bool_t
 	 * be the proxy, and checking key existence against the proxy is not
 	 * required (or sensible, as the keys may be fully virtual).
 	 */
-	duk_get_prop_stridx_short(ctx, -1, DUK_STRIDX_INT_TARGET);
-	enum_target = duk_require_hobject(ctx, -1);
+	duk_xget_owndataprop_stridx_short(thr, -1, DUK_STRIDX_INT_TARGET);
+	enum_target = duk_require_hobject(thr, -1);
 	DUK_ASSERT(enum_target != NULL);
 #if defined(DUK_USE_ES6_PROXY)
-	check_existence = (!DUK_HOBJECT_HAS_EXOTIC_PROXYOBJ(enum_target));
+	check_existence = (!DUK_HOBJECT_IS_PROXY(enum_target));
 #else
 	check_existence = 1;
 #endif
-	duk_pop(ctx);  /* still reachable */
+	duk_pop(thr);  /* still reachable */
 
 	DUK_DDD(DUK_DDDPRINT("getting next enum value, enum_target=%!iO, enumerator=%!iT",
-	                     (duk_heaphdr *) enum_target, (duk_tval *) duk_get_tval(ctx, -1)));
+	                     (duk_heaphdr *) enum_target, (duk_tval *) duk_get_tval(thr, -1)));
 
 	/* no array part */
 	for (;;) {
@@ -688,53 +625,50 @@ DUK_INTERNAL duk_bool_t duk_hobject_enumerator_next(duk_context *ctx, duk_bool_t
 
 	DUK_DDD(DUK_DDDPRINT("enumeration: updating next index to %ld", (long) idx));
 
-	duk_push_u32(ctx, (duk_uint32_t) idx);
-	duk_put_prop_stridx_short(ctx, -2, DUK_STRIDX_INT_NEXT);
+	duk_push_u32(thr, (duk_uint32_t) idx);
+	duk_put_prop_stridx_short(thr, -2, DUK_STRIDX_INT_NEXT);
 
 	/* [... enum] */
 
 	if (res) {
-		duk_push_hstring(ctx, res);
+		duk_push_hstring(thr, res);
 		if (get_value) {
-			duk_push_hobject(ctx, enum_target);
-			duk_dup_m2(ctx);       /* -> [... enum key enum_target key] */
-			duk_get_prop(ctx, -2); /* -> [... enum key enum_target val] */
-			duk_remove_m2(ctx);    /* -> [... enum key val] */
-			duk_remove(ctx, -3);   /* -> [... key val] */
+			duk_push_hobject(thr, enum_target);
+			duk_dup_m2(thr);       /* -> [... enum key enum_target key] */
+			duk_get_prop(thr, -2); /* -> [... enum key enum_target val] */
+			duk_remove_m2(thr);    /* -> [... enum key val] */
+			duk_remove(thr, -3);   /* -> [... key val] */
 		} else {
-			duk_remove_m2(ctx);    /* -> [... key] */
+			duk_remove_m2(thr);    /* -> [... key] */
 		}
 		return 1;
 	} else {
-		duk_pop(ctx);  /* -> [...] */
+		duk_pop(thr);  /* -> [...] */
 		return 0;
 	}
 }
 
 /*
- *  Get enumerated keys in an Ecmascript array.  Matches Object.keys() behavior
+ *  Get enumerated keys in an ECMAScript array.  Matches Object.keys() behavior
  *  described in E5 Section 15.2.3.14.
  */
 
-DUK_INTERNAL duk_ret_t duk_hobject_get_enumerated_keys(duk_context *ctx, duk_small_uint_t enum_flags) {
-	duk_hthread *thr = (duk_hthread *) ctx;
+DUK_INTERNAL duk_ret_t duk_hobject_get_enumerated_keys(duk_hthread *thr, duk_small_uint_t enum_flags) {
 	duk_hobject *e;
-	duk_harray *a;
 	duk_hstring **keys;
 	duk_tval *tv;
 	duk_uint_fast32_t count;
 
-	DUK_ASSERT(ctx != NULL);
-	DUK_ASSERT(duk_get_hobject(ctx, -1) != NULL);
-	DUK_UNREF(thr);
+	DUK_ASSERT(thr != NULL);
+	DUK_ASSERT(duk_get_hobject(thr, -1) != NULL);
 
 	/* Create a temporary enumerator to get the (non-duplicated) key list;
 	 * the enumerator state is initialized without being needed, but that
 	 * has little impact.
 	 */
 
-	duk_hobject_enumerator_create(ctx, enum_flags);
-	e = duk_known_hobject(ctx, -1);
+	duk_hobject_enumerator_create(thr, enum_flags);
+	e = duk_known_hobject(thr, -1);
 
 	/* [enum_target enum res] */
 
@@ -742,11 +676,10 @@ DUK_INTERNAL duk_ret_t duk_hobject_get_enumerated_keys(duk_context *ctx, duk_sma
 	DUK_ASSERT(DUK_HOBJECT_GET_ENEXT(e) >= DUK__ENUM_START_INDEX);
 	count = (duk_uint32_t) (DUK_HOBJECT_GET_ENEXT(e) - DUK__ENUM_START_INDEX);
 
-	a = duk_push_harray_with_size(ctx, count);
-	DUK_ASSERT(a != NULL);
-	DUK_ASSERT(DUK_HOBJECT_GET_ASIZE((duk_hobject *) a) == count);
-	DUK_ASSERT(a->length == count);
-	tv = DUK_HOBJECT_A_GET_BASE(thr->heap, (duk_hobject *) a);
+	/* XXX: uninit would be OK */
+	tv = duk_push_harray_with_size_outptr(thr, (duk_uint32_t) count);
+	DUK_ASSERT(count == 0 || tv != NULL);
+	DUK_ASSERT(!duk_is_bare_object(thr, -1));
 
 	/* Fill result array, no side effects. */
 
@@ -765,7 +698,7 @@ DUK_INTERNAL duk_ret_t duk_hobject_get_enumerated_keys(duk_context *ctx, duk_sma
 	}
 
 	/* [enum_target enum res] */
-	duk_remove_m2(ctx);
+	duk_remove_m2(thr);
 
 	/* [enum_target res] */
 
