@@ -24,20 +24,20 @@
  *  Dump/load helpers, xxx_raw() helpers do no buffer checks
  */
 
-DUK_LOCAL duk_uint8_t *duk__load_string_raw(duk_hthread *thr, duk_uint8_t *p) {
+DUK_LOCAL const duk_uint8_t *duk__load_string_raw(duk_hthread *thr, const duk_uint8_t *p) {
 	duk_uint32_t len;
 
-	len = DUK_RAW_READ_U32_BE(p);
+	len = DUK_RAW_READINC_U32_BE(p);
 	duk_push_lstring(thr, (const char *) p, len);
 	p += len;
 	return p;
 }
 
-DUK_LOCAL duk_uint8_t *duk__load_buffer_raw(duk_hthread *thr, duk_uint8_t *p) {
+DUK_LOCAL const duk_uint8_t *duk__load_buffer_raw(duk_hthread *thr, const duk_uint8_t *p) {
 	duk_uint32_t len;
 	duk_uint8_t *buf;
 
-	len = DUK_RAW_READ_U32_BE(p);
+	len = DUK_RAW_READINC_U32_BE(p);
 	buf = (duk_uint8_t *) duk_push_fixed_buffer_nozero(thr, (duk_size_t) len);
 	DUK_ASSERT(buf != NULL);
 	duk_memcpy((void *) buf, (const void *) p, (size_t) len);
@@ -54,7 +54,7 @@ DUK_LOCAL duk_uint8_t *duk__dump_hstring_raw(duk_uint8_t *p, duk_hstring *h) {
 	len = DUK_HSTRING_GET_BYTELEN(h);
 	DUK_ASSERT(len <= 0xffffffffUL);  /* string limits */
 	tmp32 = (duk_uint32_t) len;
-	DUK_RAW_WRITE_U32_BE(p, tmp32);
+	DUK_RAW_WRITEINC_U32_BE(p, tmp32);
 	duk_memcpy((void *) p,
 	           (const void *) DUK_HSTRING_GET_DATA(h),
 	           len);
@@ -73,7 +73,7 @@ DUK_LOCAL duk_uint8_t *duk__dump_hbuffer_raw(duk_hthread *thr, duk_uint8_t *p, d
 	len = DUK_HBUFFER_GET_SIZE(h);
 	DUK_ASSERT(len <= 0xffffffffUL);  /* buffer limits */
 	tmp32 = (duk_uint32_t) len;
-	DUK_RAW_WRITE_U32_BE(p, tmp32);
+	DUK_RAW_WRITEINC_U32_BE(p, tmp32);
 	/* When len == 0, buffer data pointer may be NULL. */
 	duk_memcpy_unsafe((void *) p,
 	                  (const void *) DUK_HBUFFER_GET_DATA_PTR(thr->heap, h),
@@ -113,7 +113,7 @@ DUK_LOCAL duk_uint8_t *duk__dump_buffer_prop(duk_hthread *thr, duk_uint8_t *p, d
 		p = duk__dump_hbuffer_raw(thr, p, h_buf);
 	} else {
 		p = DUK_BW_ENSURE_RAW(thr, bw_ctx, 4U, p);
-		DUK_RAW_WRITE_U32_BE(p, 0);
+		DUK_RAW_WRITEINC_U32_BE(p, 0);
 	}
 	return p;
 }
@@ -129,7 +129,7 @@ DUK_LOCAL duk_uint8_t *duk__dump_uint32_prop(duk_hthread *thr, duk_uint8_t *p, d
 		val = def_value;
 	}
 	p = DUK_BW_ENSURE_RAW(thr, bw_ctx, 4U, p);
-	DUK_RAW_WRITE_U32_BE(p, val);
+	DUK_RAW_WRITEINC_U32_BE(p, val);
 	return p;
 }
 
@@ -167,11 +167,11 @@ DUK_LOCAL duk_uint8_t *duk__dump_varmap(duk_hthread *thr, duk_uint8_t *p, duk_bu
 			DUK_ASSERT(DUK_HSTRING_MAX_BYTELEN <= 0x7fffffffUL);  /* ensures no overflow */
 			p = DUK_BW_ENSURE_RAW(thr, bw_ctx, 4U + DUK_HSTRING_GET_BYTELEN(key) + 4U, p);
 			p = duk__dump_hstring_raw(p, key);
-			DUK_RAW_WRITE_U32_BE(p, val);
+			DUK_RAW_WRITEINC_U32_BE(p, val);
 		}
 	}
 	p = DUK_BW_ENSURE_RAW(thr, bw_ctx, 4U, p);
-	DUK_RAW_WRITE_U32_BE(p, 0);  /* end of _Varmap */
+	DUK_RAW_WRITEINC_U32_BE(p, 0);  /* end of _Varmap */
 	return p;
 }
 
@@ -190,7 +190,7 @@ DUK_LOCAL duk_uint8_t *duk__dump_formals(duk_hthread *thr, duk_uint8_t *p, duk_b
 
 		p = DUK_BW_ENSURE_RAW(thr, bw_ctx, 4U, p);
 		DUK_ASSERT(h->length != DUK__NO_FORMALS);  /* limits */
-		DUK_RAW_WRITE_U32_BE(p, h->length);
+		DUK_RAW_WRITEINC_U32_BE(p, h->length);
 
 		for (i = 0; i < h->length; i++) {
 			duk_tval *tv_val;
@@ -211,7 +211,7 @@ DUK_LOCAL duk_uint8_t *duk__dump_formals(duk_hthread *thr, duk_uint8_t *p, duk_b
 	} else {
 		DUK_DD(DUK_DDPRINT("dumping function without _Formals, emit marker to indicate missing _Formals"));
 		p = DUK_BW_ENSURE_RAW(thr, bw_ctx, 4U, p);
-		DUK_RAW_WRITE_U32_BE(p, DUK__NO_FORMALS);  /* marker: no formals */
+		DUK_RAW_WRITEINC_U32_BE(p, DUK__NO_FORMALS);  /* marker: no formals */
 	}
 	return p;
 }
@@ -251,27 +251,27 @@ static duk_uint8_t *duk__dump_func(duk_hthread *thr, duk_hcompfunc *func, duk_bu
 
 	/* Fixed header info. */
 	tmp32 = count_instr;
-	DUK_RAW_WRITE_U32_BE(p, tmp32);
+	DUK_RAW_WRITEINC_U32_BE(p, tmp32);
 	tmp32 = (duk_uint32_t) DUK_HCOMPFUNC_GET_CONSTS_COUNT(thr->heap, func);
-	DUK_RAW_WRITE_U32_BE(p, tmp32);
+	DUK_RAW_WRITEINC_U32_BE(p, tmp32);
 	tmp32 = (duk_uint32_t) DUK_HCOMPFUNC_GET_FUNCS_COUNT(thr->heap, func);
-	DUK_RAW_WRITE_U32_BE(p, tmp32);
+	DUK_RAW_WRITEINC_U32_BE(p, tmp32);
 	tmp16 = func->nregs;
-	DUK_RAW_WRITE_U16_BE(p, tmp16);
+	DUK_RAW_WRITEINC_U16_BE(p, tmp16);
 	tmp16 = func->nargs;
-	DUK_RAW_WRITE_U16_BE(p, tmp16);
+	DUK_RAW_WRITEINC_U16_BE(p, tmp16);
 #if defined(DUK_USE_DEBUGGER_SUPPORT)
 	tmp32 = func->start_line;
-	DUK_RAW_WRITE_U32_BE(p, tmp32);
+	DUK_RAW_WRITEINC_U32_BE(p, tmp32);
 	tmp32 = func->end_line;
-	DUK_RAW_WRITE_U32_BE(p, tmp32);
+	DUK_RAW_WRITEINC_U32_BE(p, tmp32);
 #else
-	DUK_RAW_WRITE_U32_BE(p, 0);
-	DUK_RAW_WRITE_U32_BE(p, 0);
+	DUK_RAW_WRITEINC_U32_BE(p, 0);
+	DUK_RAW_WRITEINC_U32_BE(p, 0);
 #endif
 	tmp32 = DUK_HEAPHDR_GET_FLAGS((duk_heaphdr *) func);  /* masks flags, only duk_hobject flags */
 	tmp32 &= ~(DUK_HOBJECT_FLAG_HAVE_FINALIZER);  /* finalizer flag is lost */
-	DUK_RAW_WRITE_U32_BE(p, tmp32);
+	DUK_RAW_WRITEINC_U32_BE(p, tmp32);
 
 	/* Bytecode instructions: endian conversion needed unless
 	 * platform is big endian.
@@ -285,7 +285,7 @@ static duk_uint8_t *duk__dump_func(duk_hthread *thr, duk_hcompfunc *func, duk_bu
 #else
 	while (ins != ins_end) {
 		tmp32 = (duk_uint32_t) (*ins);
-		DUK_RAW_WRITE_U32_BE(p, tmp32);
+		DUK_RAW_WRITEINC_U32_BE(p, tmp32);
 		ins++;
 	}
 #endif
@@ -310,7 +310,7 @@ static duk_uint8_t *duk__dump_func(duk_hthread *thr, duk_hcompfunc *func, duk_bu
 			p = DUK_BW_ENSURE_RAW(thr, bw_ctx, 1U + 8U, p);
 			*p++ = DUK__SER_NUMBER;
 			d = DUK_TVAL_GET_NUMBER(tv);
-			DUK_RAW_WRITE_DOUBLE_BE(p, d);
+			DUK_RAW_WRITEINC_DOUBLE_BE(p, d);
 		}
 		tv++;
 	}
@@ -371,7 +371,7 @@ static duk_uint8_t *duk__dump_func(duk_hthread *thr, duk_hcompfunc *func, duk_bu
 		DUK_ASSERT((duk_size_t) (p_end - p) >= (duk_size_t) (n)); \
 	} while (0)
 
-static duk_uint8_t *duk__load_func(duk_hthread *thr, duk_uint8_t *p, duk_uint8_t *p_end) {
+static const duk_uint8_t *duk__load_func(duk_hthread *thr, const duk_uint8_t *p, const duk_uint8_t *p_end) {
 	duk_hcompfunc *h_fun;
 	duk_hbuffer *h_data;
 	duk_size_t data_size;
@@ -395,12 +395,12 @@ static duk_uint8_t *duk__load_func(duk_hthread *thr, duk_uint8_t *p, duk_uint8_t
 
 	DUK_ASSERT(thr != NULL);
 
-	DUK_DD(DUK_DDPRINT("loading function, p=%p, p_end=%p", (void *) p, (void *) p_end));
+	DUK_DD(DUK_DDPRINT("loading function, p=%p, p_end=%p", (const void *) p, (const void *) p_end));
 
 	DUK__ASSERT_LEFT(3 * 4);
-	count_instr = DUK_RAW_READ_U32_BE(p);
-	count_const = DUK_RAW_READ_U32_BE(p);
-	count_funcs = DUK_RAW_READ_U32_BE(p);
+	count_instr = DUK_RAW_READINC_U32_BE(p);
+	count_const = DUK_RAW_READINC_U32_BE(p);
+	count_funcs = DUK_RAW_READINC_U32_BE(p);
 
 	data_size = sizeof(duk_tval) * count_const +
 	            sizeof(duk_hobject *) * count_funcs +
@@ -428,17 +428,17 @@ static duk_uint8_t *duk__load_func(duk_hthread *thr, duk_uint8_t *p, duk_uint8_t
 	DUK_ASSERT(DUK_HCOMPFUNC_GET_BYTECODE(thr->heap, h_fun) == NULL);
 	DUK_ASSERT(DUK_HOBJECT_GET_PROTOTYPE(thr->heap, (duk_hobject *) h_fun) == thr->builtins[DUK_BIDX_FUNCTION_PROTOTYPE]);
 
-	h_fun->nregs = DUK_RAW_READ_U16_BE(p);
-	h_fun->nargs = DUK_RAW_READ_U16_BE(p);
+	h_fun->nregs = DUK_RAW_READINC_U16_BE(p);
+	h_fun->nargs = DUK_RAW_READINC_U16_BE(p);
 #if defined(DUK_USE_DEBUGGER_SUPPORT)
-	h_fun->start_line = DUK_RAW_READ_U32_BE(p);
-	h_fun->end_line = DUK_RAW_READ_U32_BE(p);
+	h_fun->start_line = DUK_RAW_READINC_U32_BE(p);
+	h_fun->end_line = DUK_RAW_READINC_U32_BE(p);
 #else
 	p += 8;  /* skip line info */
 #endif
 
 	/* duk_hcompfunc flags; quite version specific */
-	tmp32 = DUK_RAW_READ_U32_BE(p);
+	tmp32 = DUK_RAW_READINC_U32_BE(p);
 	DUK_HEAPHDR_SET_FLAGS((duk_heaphdr *) h_fun, tmp32);  /* masks flags to only change duk_hobject flags */
 
 	/* standard prototype (no need to set here, already set) */
@@ -474,7 +474,7 @@ static duk_uint8_t *duk__load_func(duk_hthread *thr, duk_uint8_t *p, duk_uint8_t
 #else
 	q = fun_data + sizeof(duk_tval) * count_const + sizeof(duk_hobject *) * count_funcs;
 	for (n = count_instr; n > 0; n--) {
-		*((duk_instr_t *) (void *) q) = DUK_RAW_READ_U32_BE(p);
+		*((duk_instr_t *) (void *) q) = DUK_RAW_READINC_U32_BE(p);
 		q += sizeof(duk_instr_t);
 	}
 #endif
@@ -482,7 +482,7 @@ static duk_uint8_t *duk__load_func(duk_hthread *thr, duk_uint8_t *p, duk_uint8_t
 	/* Load constants onto value stack but don't yet copy to buffer. */
 	for (n = count_const; n > 0; n--) {
 		DUK__ASSERT_LEFT(1);
-		const_type = DUK_RAW_READ_U8(p);
+		const_type = DUK_RAW_READINC_U8(p);
 		switch (const_type) {
 		case DUK__SER_STRING: {
 			p = duk__load_string_raw(thr, p);
@@ -495,7 +495,7 @@ static duk_uint8_t *duk__load_func(duk_hthread *thr, duk_uint8_t *p, duk_uint8_t
 			duk_tval tv_tmp;
 			duk_double_t val;
 			DUK__ASSERT_LEFT(8);
-			val = DUK_RAW_READ_DOUBLE_BE(p);
+			val = DUK_RAW_READINC_DOUBLE_BE(p);
 			DUK_TVAL_SET_NUMBER_CHKFAST_SLOW(&tv_tmp, val);
 			duk_push_tval(thr, &tv_tmp);
 			break;
@@ -562,7 +562,7 @@ static duk_uint8_t *duk__load_func(duk_hthread *thr, duk_uint8_t *p, duk_uint8_t
 	duk_set_top(thr, idx_base + 1);
 
 	/* Setup function properties. */
-	tmp32 = DUK_RAW_READ_U32_BE(p);
+	tmp32 = DUK_RAW_READINC_U32_BE(p);
 	duk_push_u32(thr, tmp32);
 	duk_xdef_prop_stridx_short(thr, -2, DUK_STRIDX_LENGTH, DUK_PROPDESC_FLAGS_C);
 
@@ -642,7 +642,7 @@ static duk_uint8_t *duk__load_func(duk_hthread *thr, duk_uint8_t *p, duk_uint8_t
 			duk_pop(thr);
 			break;
 		}
-		tmp32 = DUK_RAW_READ_U32_BE(p);
+		tmp32 = DUK_RAW_READINC_U32_BE(p);
 		duk_push_u32(thr, tmp32);
 		duk_put_prop(thr, -3);
 	}
@@ -652,7 +652,7 @@ static duk_uint8_t *duk__load_func(duk_hthread *thr, duk_uint8_t *p, duk_uint8_t
 	/* _Formals may have been missing in the original function, which is
 	 * handled using a marker length.
 	 */
-	arr_limit = DUK_RAW_READ_U32_BE(p);
+	arr_limit = DUK_RAW_READINC_U32_BE(p);
 	if (arr_limit != DUK__NO_FORMALS) {
 		duk_push_bare_array(thr);  /* _Formals */
 		for (arr_idx = 0; arr_idx < arr_limit; arr_idx++) {
@@ -706,7 +706,7 @@ DUK_EXTERNAL void duk_dump_function(duk_hthread *thr) {
 }
 
 DUK_EXTERNAL void duk_load_function(duk_hthread *thr) {
-	duk_uint8_t *p_buf, *p, *p_end;
+	const duk_uint8_t *p_buf, *p, *p_end;
 	duk_size_t sz;
 
 	DUK_ASSERT_API_ENTRY(thr);
